/*
      SoundMan v0.1 - Realtime frequency analyzer
      Copyright (C) 1999 Gabor Kerenyi (wom@fireball.kalmar.hu)
*/

#include <stdio.h>
#include <math.h>
#include <qwidget.h>
#include "fft.h"
#include "main.h"
#include <fcntl.h>
#include <qapplication.h>

extern cnf mconf;

int n,logn;

#define PI 3.141592


complex *source;
complex *DFT;
complex *exp_tabl; 
complex *fftbuff; 


Fft::Fft(QWidget *parent, const char *name): QWidget(parent,name)
{
logn=0;
n=0;
}

void Fft::cp_datas(int fr,int to,unsigned char *wave)
{
   int i,k;
   double j,tmp;
   for (i=fr;i<to;i++)
   {
       tmp=*(wave+i);
       j=i-fr;
       k=(int)j;
       switch(mconf.spect.window)
	{
	    case 1:
		tmp*=1-fabs((double)n/2-j)/((double)n/2);
		break;
	    case 2:
		tmp*=0.5-0.5*cos(2*PI*j/(double)n);
		break;
	    case 3:
		tmp*=0.54-0.46*cos(2*PI*j/(double)n);
		break;
	}
	(mconf.spect.source+k)->re=tmp;
	(mconf.spect.source+k)->im=0;
   }
} 


void Fft::init(long mln)
{
long mn;
 mn=1<<mln;
 if (mln!=logn){
	if (mconf.spect.power!=NULL)
	   free(mconf.spect.power);
	if (mconf.spect.phase!=NULL)
	   free(mconf.spect.phase);
	DFT=(complex *)calloc(mn,sizeof(complex));
	mconf.spect.power=(double *)calloc(mn,sizeof(double));
	mconf.spect.phase=(double *)calloc(mn,sizeof(double));
 }
  logn=mln;
  n=1<<mln;
}

void cx_mul(complex *a,complex *b, complex *dest)
{ 
dest->re= a->re*b->re - a->im*b->im;
dest->im= a->re*b->im + b->re*a->im;
}

void cx_copy(complex *a,complex *b)
{  
  b->re=a->re;
  b->im=a->im;
}

void cx_add(complex *a,complex *b, complex *dest)
{ 
dest->re= a->re + b->re;
dest->im= a->im + b->im;
}

void cx_sub(complex *a,complex *b, complex *dest)
{
dest->re= a->re - b->re;
dest->im= a->im - b->im;
}


void Fft::dft(int wave_start)
{
  int N,k;
  complex tmp3;
  double konst,konst2;
  konst=(double)2*PI/(double)n;
  for (k=0;k<n;k++)
  {
     tmp3.re=0;
     tmp3.im=0;
     qApp->processEvents();
     konst2=konst*(double)k;
     for (N=0;N<n;N++)
     {
	tmp3.re+=(mconf.spect.source+N)->re*cos(konst2*(double)N);
	tmp3.im+=(mconf.spect.source+N)->im*sin(konst2*(double)N);
     }
     if ((k>64) && (k<8192-64)){
        *(mconf.wave.wave+(8192-k)+wave_start)=(unsigned char)(tmp3.re+tmp3.im);
     }

  }
}

void Fft::fft()
{
int p,q,p2,r,i,nper2p;
float s,s1;
complex *p_from,*p_to,*p1_from,*p1_to,*p2_from,*p2_to;
complex cx,cxexp;

source=mconf.spect.source;
p_from=mconf.spect.source;
p_to=DFT;

p=n;
for (i=0;i<logn;i++)
{
  p2=p;
  p>>=1;
  nper2p=n/p2;
  p2_from=p+(p1_from=p_from);
  p2_to=n/2+(p1_to=p_to);

  s=(float)p*2.0*PI/(float)n;
  s1=0;
  for (r=0;r<nper2p;r++)
  {
     cxexp.re=cos(s1);
     cxexp.im=-sin(s1);
     for (q=0;q<p;q++)
     {
	cx_mul(p2_from,&cxexp,&cx);
	cx_add(p1_from,&cx,p1_to);
	cx_sub(p1_from,&cx,p2_to);
	p1_from++;
	p2_from++;
	p1_to++;
	p2_to++;
     }
     s1+=s;
     p1_from+=p;
     p2_from+=p;
  }
  p1_from=p_from;
  p_from=p_to;
  p_to=p1_from;
}  
if (logn%2==1)
{
  p1_from=DFT;
  p1_to=mconf.spect.source;
  for (i=0;i<n;i++)
  {
    *p1_to=*p1_from;
    p1_from++;
    p1_to++;
  }
}
mconf.spect.maxpower=0;
for (i=0;i<n;i++)
{
  (mconf.spect.source+i)->re/=n;
  (mconf.spect.source+i)->im/=n;
  *(mconf.spect.power+i)=sqrt((source+i)->re*(source+i)->re+(source+i)->im*(source+i)->im);
  if (i>1) if (*(mconf.spect.power+i)>mconf.spect.maxpower) mconf.spect.maxpower=*(mconf.spect.power+i);

}      	    
mconf.spect.pn=n/2;
}

Fft::~Fft()
{
if (DFT!=NULL) free(DFT);
}

